from typing import Iterable, TypeVar, Generic, Tuple, Optional
from collections.abc import Sequence

T = TypeVar('T')

__all__ = [
    'RunLengthEncoder'
]


class RunLengthEncoder(Sequence, Generic[T]):

    def __init__(self, collection: Optional[Iterable[T]] = None):
        Sequence.__init__(self)
        self._values: list[T] = []
        self._multiplicity: list[int] = []

        if collection is not None:
            self.extend(collection)

    def append(self, value: T) -> None:
        if len(self) == 0 or self._values[-1] != value:
            self._values.append(value)
            self._multiplicity.append(1)
            return

        self._multiplicity[-1] += 1

    def extend(self, collection: Iterable[T]) -> None:
        for item in collection:
            self.append(item)

    @property
    def values(self) -> Sequence[T]:
        return tuple(self._values)

    @property
    def multiplicity(self) -> Sequence[int]:
        return tuple(self._multiplicity)

    def __getitem__(self, i) -> Tuple[T, int]:
        return (
            self._values.__getitem__(i),
            self._multiplicity.__getitem__(i))

    def __len__(self) -> int:
        return sum(self._multiplicity)

    def __iadd__(self, other):
        self.append(other)
        return self

    def __repr__(self):
        return f'{self.__class__.__name__}({list(self)})'

